import { registerLocaleData } from '@angular/common';
import localeDe from '@angular/common/locales/de';
import localeDeExtra from '@angular/common/locales/extra/de';
import { VorgangBescheidWizardE2EComponent } from 'apps/alfa-e2e/src/components/vorgang/vorgang-bescheid-wizard.e2e.component';
import { VorgangFormularButtonsE2EComponent } from 'apps/alfa-e2e/src/components/vorgang/vorgang-formular-buttons.e2e.components';
import { VorgangE2E, VorgangStatusE2E } from 'apps/alfa-e2e/src/model/vorgang';
import { uploadFile } from 'apps/alfa-e2e/src/support/file-upload';
import 'cypress-real-events/support';
import { VorgangListE2EComponent } from '../../../components/vorgang/vorgang-list.e2e.component';
import { MainPage, waitForSpinnerToDisappear } from '../../../page-objects/main.po';
import { VorgangPage } from '../../../page-objects/vorgang.po';
import { dropCollections, getTestElement } from '../../../support/cypress-helper';
import { contains, exist, haveValue, notContains, notExist } from '../../../support/cypress.util';
import {
  TEST_FILE_BESCHEID_ANHANG_BIG,
  TEST_FILE_BESCHEID_ANHANG_VALID,
  TEST_FILE_BESCHEID_VALID,
  TEST_FILE_JPEG,
  TEST_FILE_JPG,
  TEST_FILE_PNG,
  TEST_FILE_WITH_CONTENT,
} from '../../../support/data.util';
import { initUsermanagerUsers, loginAsSabine } from '../../../support/user-util';
import { buildVorgang, initVorgaenge, objectIds } from '../../../support/vorgang-util';

registerLocaleData(localeDe, 'de', localeDeExtra);

describe('Bescheid Dokumente hochladen', () => {
  const mainPage: MainPage = new MainPage();
  const vorgangList: VorgangListE2EComponent = mainPage.getVorgangList();

  const vorgangPage: VorgangPage = new VorgangPage();
  const bescheidWizard: VorgangBescheidWizardE2EComponent = vorgangPage.getBescheidWizard();

  const bescheidVorgang: VorgangE2E = {
    ...buildVorgang(objectIds[0], 'Wizard Vorgang'),
    status: VorgangStatusE2E.IN_BEARBEITUNG,
  };

  const vorgangFormularButtons: VorgangFormularButtonsE2EComponent =
    vorgangPage.getFormularButtons();

  const documentError: string = 'Erlaubte Dateiendungen';
  const sizeError: string = 'Anhänge größer';
  const missingBescheidError: string = 'Bitte fügen Sie';
  const nachrichtHeader: string = 'Neue Nachricht';
  const antragstellerName: string = 'An: Max Testermann';
  const betreffText: string = 'Ihr Bescheid zum Antrag';
  const nachrichtText: string =
    'Sehr geehrte/r Antragsteller/in,\n\nim Folgenden erhalten Sie Ihren Bescheid.\n\nMit freundlichen Grüßen\n\nIhre Verwaltung';

  before(() => {
    initVorgaenge([bescheidVorgang]);
    initUsermanagerUsers();

    loginAsSabine();

    waitForSpinnerToDisappear();
    exist(vorgangList.getRoot());
  });

  after(() => {
    dropCollections();
  });

  describe('Upload Bescheid Document', () => {
    it('should open Bescheid Wizard', () => {
      vorgangList.getListItem(bescheidVorgang.name).getRoot().click();
      waitForSpinnerToDisappear();

      vorgangFormularButtons.getBescheidenButton().click();
      exist(bescheidWizard.getRoot());

      bescheidWizard.getWeiterButton().click();
    });

    it('should upload manual Bescheid file', () => {
      uploadFile(bescheidWizard.getUploadBescheidButton(), TEST_FILE_BESCHEID_VALID);
      notExist(bescheidWizard.getBescheidUploadSpinner());
      exist(bescheidWizard.getFileBescheidValidInWizard());
    });

    it('should upload attachment file', () => {
      uploadFile(bescheidWizard.getUploadAttachmentButton(), TEST_FILE_BESCHEID_ANHANG_VALID);
      exist(bescheidWizard.getFileAnhangValidInWizard());
    });

    it('should still show files after clicking Weiter and step 2', () => {
      bescheidWizard.getWeiterButton().click();
      bescheidWizard.getStepButton2().click();
      exist(bescheidWizard.getFileBescheidValidInWizard());
      exist(bescheidWizard.getFileAnhangValidInWizard());
    });

    it('should still show file after closing the Wizard and continue working', () => {
      bescheidWizard.getCloseButton().click();
      bescheidWizard.getCloseSpeichernButton().click();
      waitForSpinnerToDisappear();

      vorgangFormularButtons.getBescheidenButton().click();
      bescheidWizard.getWeiterButton().click();
      exist(bescheidWizard.getFileBescheidValidInWizard());
      exist(bescheidWizard.getFileAnhangValidInWizard());
    });

    it('should delete files on clicking X button', () => {
      bescheidWizard
        .getDeleteButtonOfElement(
          bescheidWizard.getElementFromFileName(TEST_FILE_BESCHEID_ANHANG_VALID),
        )
        .click();
      notExist(bescheidWizard.getFileAnhangValidInWizard());
      exist(bescheidWizard.getFileBescheidValidInWizard());

      bescheidWizard
        .getDeleteButtonOfElement(bescheidWizard.getElementFromFileName(TEST_FILE_BESCHEID_VALID))
        .click();
      notExist(bescheidWizard.getFileBescheidValidInWizard());
    });

    it('should be able to attach all valid types of files', () => {
      uploadFile(bescheidWizard.getUploadAttachmentButton(), TEST_FILE_JPG);
      notExist(bescheidWizard.getAttachmentUploadSpinner());
      uploadFile(bescheidWizard.getUploadAttachmentButton(), TEST_FILE_JPEG);
      notExist(bescheidWizard.getAttachmentUploadSpinner());
      uploadFile(bescheidWizard.getUploadAttachmentButton(), TEST_FILE_PNG);
      notExist(bescheidWizard.getAttachmentUploadSpinner());
      exist(getTestElement(bescheidWizard.getElementFromFileName(TEST_FILE_JPG)));
      exist(getTestElement(bescheidWizard.getElementFromFileName(TEST_FILE_JPEG)));
      exist(getTestElement(bescheidWizard.getElementFromFileName(TEST_FILE_PNG)));
    });
  });

  describe('File upload error handling', () => {
    it('should show error for invalid file type', () => {
      uploadFile(bescheidWizard.getUploadBescheidButton(), TEST_FILE_WITH_CONTENT);
      contains(bescheidWizard.getBescheidDocument(), documentError);
    });

    it('should show error if file is too big', () => {
      uploadFile(bescheidWizard.getUploadAttachmentButton(), TEST_FILE_BESCHEID_ANHANG_BIG);
      contains(bescheidWizard.getAttachmentDocument(), sizeError);
    });
  });

  describe('do not continue without attached Bescheid', () => {
    it('should show error if no Bescheid is attached and step 3 is clicked', () => {
      bescheidWizard.getWeiterButton().click();
      contains(bescheidWizard.getBescheidDocument(), missingBescheidError);
    });
  });

  describe('check contents of step 3', () => {
    it('should show Max Testermann as Antragsteller, the default message text', () => {
      uploadFile(bescheidWizard.getUploadBescheidButton(), TEST_FILE_BESCHEID_VALID);
      notExist(bescheidWizard.getBescheidUploadSpinner());

      bescheidWizard.getWeiterButton().click();

      contains(bescheidWizard.getAntragstellerText(), nachrichtHeader);
      contains(bescheidWizard.getAntragstellerText(), antragstellerName);
      haveValue(bescheidWizard.getBetreffText(), betreffText);
      haveValue(bescheidWizard.getNachrichtText(), nachrichtText);
    });

    it('should not contain error message from upload', () => {
      notContains(bescheidWizard.getAttachmentDocument(), sizeError);
    });
  });
});
