import {
  BinaryFileListLinkRel,
  BinaryFileResource,
  BinaryFileService,
} from '@alfa-client/binary-file-shared';
import {
  CommandOrder,
  CommandResource,
  CommandResourceService,
  CommandService,
  getEffectedResourceUrl,
  notHasCommandError,
  tapOnCommandSuccessfullyDone,
} from '@alfa-client/command-shared';
import {
  createEmptyStateResource,
  createStateResource,
  EMPTY_ARRAY,
  filterIsLoadedOrHasError,
  getEmbeddedResources,
  hasStateResourceError,
  HttpError,
  isLoaded,
  isNotEmpty,
  isNotNil,
  ResourceListService,
  sortByGermanDateStr,
  StateResource,
} from '@alfa-client/tech-shared';
import {
  VorgangCommandService,
  VorgangService,
  VorgangWithEingangLinkRel,
  VorgangWithEingangResource,
} from '@alfa-client/vorgang-shared';
import { getEmpfaenger } from '@alfa-client/vorgang-shared-ui';
import { Injectable } from '@angular/core';
import { getUrl, hasLink, ResourceUri } from '@ngxp/rest';
import {
  BehaviorSubject,
  filter,
  first,
  map,
  Observable,
  startWith,
  Subscription,
  switchMap,
  take,
  tap,
} from 'rxjs';
import {
  ListResourceServiceConfig,
  ResourceServiceConfig,
} from '../../../tech-shared/src/lib/resource/resource.model';
import { ResourceRepository } from '../../../tech-shared/src/lib/resource/resource.repository';
import { ResourceService } from '../../../tech-shared/src/lib/resource/resource.service';
import { BescheidFacade } from './+state/bescheid.facade';
import { BescheidLinkRel, BescheidListLinkRel } from './bescheid.linkrel';
import {
  Bescheid,
  BescheidListResource,
  BescheidResource,
  BescheidStatus,
  UploadFileInProgress,
} from './bescheid.model';
import {
  buildCreateBescheidCommand,
  buildCreateBescheidDocumentCommandProps,
  buildCreateBescheidDocumentFromFileProps,
  buildDeleteBescheidCommandProps,
  buildSendBescheidCommandProps,
  buildUpdateBescheidCommandProps,
} from './bescheid.util';
import { DocumentLinkRel } from './document.linkrel';
import { DocumentResource } from './document.model';

@Injectable({ providedIn: 'root' })
export class BescheidService {
  bescheidResourceService: ResourceService<VorgangWithEingangResource, BescheidResource>;
  bescheidListResourceService: ResourceListService<
    VorgangWithEingangResource,
    BescheidListResource,
    BescheidResource
  >;

  readonly bescheidDocumentFile$: BehaviorSubject<StateResource<BinaryFileResource>> =
    new BehaviorSubject<StateResource<BinaryFileResource>>(createEmptyStateResource());

  readonly bescheidDocumentUri$: BehaviorSubject<ResourceUri> = new BehaviorSubject<ResourceUri>(
    null,
  );

  readonly bescheidDocument$: BehaviorSubject<StateResource<DocumentResource>> =
    new BehaviorSubject<StateResource<DocumentResource>>(createEmptyStateResource());

  readonly bescheidList$: BehaviorSubject<StateResource<BescheidListResource>> =
    new BehaviorSubject(createEmptyStateResource<BescheidListResource>());

  readonly createBescheidDocumentInProgress$: BehaviorSubject<StateResource<CommandResource>> =
    new BehaviorSubject<StateResource<CommandResource>>(createEmptyStateResource());

  readonly uploadBescheidDocumentInProgress$: BehaviorSubject<UploadFileInProgress> =
    new BehaviorSubject<UploadFileInProgress>({ loading: false });

  readonly uploadAttachmentInProgress$: BehaviorSubject<UploadFileInProgress> = new BehaviorSubject(
    { loading: false },
  );

  readonly uploadedAttachment$: BehaviorSubject<StateResource<BinaryFileResource>> =
    new BehaviorSubject(createEmptyStateResource());

  loadBescheidDocumentSubscription: Subscription;

  constructor(
    private readonly facade: BescheidFacade,
    private readonly vorgangService: VorgangService,
    private readonly commandService: CommandService,
    private readonly vorgangCommandService: VorgangCommandService,
    private readonly binaryFileService: BinaryFileService,
    private readonly repository: ResourceRepository,
  ) {
    this.bescheidResourceService = new CommandResourceService(
      this.buildBescheidDraftServiceConfig(),
      repository,
      this.commandService,
    );
    this.bescheidListResourceService = new ResourceListService(
      this.buildBescheidListServiceConfig(),
      repository,
    );
  }

  buildBescheidDraftServiceConfig(): ResourceServiceConfig<VorgangWithEingangResource> {
    return {
      resource: this.vorgangService.getVorgangWithEingang(),
      getLinkRel: VorgangWithEingangLinkRel.BESCHEID_DRAFT,
      delete: { linkRel: BescheidLinkRel.DELETE, order: CommandOrder.DELETE_BESCHEID },
      edit: { linkRel: BescheidLinkRel.UPDATE, order: CommandOrder.UPDATE_BESCHEID },
    };
  }

  buildBescheidListServiceConfig(): ListResourceServiceConfig<VorgangWithEingangResource> {
    return {
      baseResource: this.vorgangService.getVorgangWithEingang(),
      listLinkRel: VorgangWithEingangLinkRel.BESCHEIDE,
      listResourceListLinkRel: BescheidListLinkRel.BESCHEID_LIST,
    };
  }

  public init(): void {
    this.bescheidResourceService = new CommandResourceService(
      this.buildBescheidDraftServiceConfig(),
      this.repository,
      this.commandService,
    );
    this.bescheidDocumentFile$.next(createEmptyStateResource());
    this.bescheidDocumentUri$.next(null);
    this.uploadBescheidDocumentInProgress$.next({ loading: false });
    this.clearUploadAttachment();
    this.uploadAttachmentInProgress$.next({ loading: false });
  }

  public getBescheidDraft(): Observable<StateResource<BescheidResource>> {
    return this.bescheidResourceService.get();
  }

  public getBescheidCommand(): Observable<StateResource<CommandResource>> {
    return this.facade.getBescheidCommand();
  }

  public createBescheid(
    vorgangWithEingang: VorgangWithEingangResource,
    bescheid?: Bescheid,
  ): Observable<StateResource<CommandResource>> {
    this.facade.createBescheidDraft(vorgangWithEingang, buildCreateBescheidCommand(bescheid));
    return this.getBescheidCommand().pipe(
      tapOnCommandSuccessfullyDone((commandStateResource: StateResource<CommandResource>) =>
        this.updateBescheidDraft(commandStateResource.resource),
      ),
    );
  }

  public bescheidErstellungUeberspringen(
    vorgangWithEingangResource: VorgangWithEingangResource,
  ): Observable<StateResource<CommandResource>> {
    if (!this.existsBescheidDraft()) {
      return this.vorgangAbschliesen(vorgangWithEingangResource);
    }
    return this.getBescheidDraft().pipe(
      filter(isLoaded),
      first(),
      switchMap((bescheidStateResource: StateResource<BescheidResource>) =>
        this.bescheidLoeschenUndErstellungUeberspringen(
          vorgangWithEingangResource,
          bescheidStateResource.resource,
        ),
      ),
    );
  }

  public getBescheidDraftIfExists(): Observable<StateResource<BescheidResource>> {
    return this.vorgangService.getVorgangWithEingang().pipe(
      filter(
        (stateResource: StateResource<VorgangWithEingangResource>) =>
          isLoaded(stateResource) &&
          hasLink(stateResource.resource, VorgangWithEingangLinkRel.BESCHEID_DRAFT),
      ),
      switchMap(() => this.getBescheidDraft()),
      startWith(createEmptyStateResource<BescheidResource>()),
    );
  }

  bescheidLoeschenUndErstellungUeberspringen(
    vorgangWithEingangResource: VorgangWithEingangResource,
    bescheidResource: BescheidResource,
  ): Observable<StateResource<CommandResource>> {
    return this.vorgangAbschliesen(vorgangWithEingangResource).pipe(
      tapOnCommandSuccessfullyDone(() => this.deleteBescheid(bescheidResource)),
    );
  }

  vorgangAbschliesen(
    vorgangWithEingangResource: VorgangWithEingangResource,
  ): Observable<StateResource<CommandResource>> {
    return this.vorgangCommandService.abschliessen(vorgangWithEingangResource);
  }

  public updateBescheid(bescheid: Bescheid): Observable<StateResource<CommandResource>> {
    return this.doUpdateBescheid(this.getResource(), bescheid).pipe(
      tapOnCommandSuccessfullyDone((commandStateResource: StateResource<CommandResource>) => {
        this.updateBescheidDraft(commandStateResource.resource);
        this.clearCreateBescheidDocumentInProgress();
        this.clearUploadBescheidDocumentInProgress();
        this.clearUploadAttachment();
      }),
    );
  }

  private clearUploadAttachment() {
    this.uploadedAttachment$.next(createEmptyStateResource());
  }

  public sendBescheidToAntragsteller(
    bescheidResource: BescheidResource,
  ): Observable<StateResource<CommandResource>> {
    return this.sendBescheid(bescheidResource, BescheidLinkRel.BESCHEIDEN_UND_SENDEN);
  }

  public sendBescheidManually(
    bescheidResource: BescheidResource,
  ): Observable<StateResource<CommandResource>> {
    return this.sendBescheid(bescheidResource, BescheidLinkRel.BESCHEIDEN);
  }

  sendBescheid(
    bescheidResource: BescheidResource,
    linkRel: string,
  ): Observable<StateResource<CommandResource>> {
    return this.bescheidResourceService.get().pipe(
      filterIsLoadedOrHasError(),
      switchMap((stateResource: StateResource<BescheidResource>) =>
        this.commandService.createCommandByProps(
          buildSendBescheidCommandProps(stateResource.resource, linkRel),
        ),
      ),
    );
  }

  doUpdateBescheid(
    bescheidResource: BescheidResource,
    bescheid: Bescheid,
  ): Observable<StateResource<CommandResource>> {
    return this.commandService.createCommandByProps(
      buildUpdateBescheidCommandProps(bescheidResource, bescheid),
    );
  }

  private updateBescheidDraft(command: CommandResource): void {
    this.bescheidResourceService.loadByResourceUri(getEffectedResourceUrl(command));
  }

  public getAttachments(): Observable<BinaryFileResource[]> {
    return this.getBescheidDraft().pipe(
      filter(isLoaded),
      map((stateResource: StateResource<BescheidResource>) => stateResource.resource),
      filter((resource: BescheidResource) => hasLink(resource, BescheidLinkRel.ATTACHMENTS)),
      switchMap((resource: BescheidResource) =>
        this.binaryFileService.getFiles(resource, BescheidLinkRel.ATTACHMENTS),
      ),
      filter(isLoaded),
      map((stateResource) =>
        getEmbeddedResources<BinaryFileResource>(stateResource, BinaryFileListLinkRel.FILE_LIST),
      ),
    );
  }

  public getDocumentUri(): Observable<ResourceUri> {
    return this.bescheidDocumentUri$.asObservable();
  }

  public setDocumentUri(uri: ResourceUri): void {
    this.bescheidDocumentUri$.next(uri);
  }

  public existBescheidDocumentFile(): Observable<boolean> {
    return this.bescheidDocumentFile$.asObservable().pipe(map(isLoaded));
  }

  public getBescheidDocumentFile(): Observable<StateResource<BinaryFileResource>> {
    return this.bescheidDocumentFile$.asObservable();
  }

  public getBescheidDocument(): Observable<StateResource<DocumentResource>> {
    return this.bescheidDocument$.asObservable();
  }

  public loadBescheidDocumentByUri(resourceUri: ResourceUri): void {
    this.setBescheidDocumentFileLoading();
    this.loadBescheidDocumentSubscription = this.repository
      .getResource(resourceUri)
      .pipe()
      .subscribe((document: DocumentResource) => {
        this.bescheidDocument$.next(createStateResource(document));
        this.loadBescheidDocumentFile(document);
        this.loadBescheidDocumentSubscription.unsubscribe();
      });
  }

  setBescheidDocumentFileLoading(): void {
    this.bescheidDocumentFile$.next({ ...this.bescheidDocumentFile$.value, loading: true });
  }

  public uploadBescheidDocument(
    bescheid: BescheidResource,
    file: File,
  ): Observable<UploadFileInProgress> {
    this.clearCreateBescheidDocumentInProgress();
    this.initUploadBescheidDocumentInProgress(file.name);
    this.doUploadBescheidDocument(bescheid, file);
    return this.getUploadBescheidDocumentInProgress();
  }

  private clearCreateBescheidDocumentInProgress(): void {
    this.createBescheidDocumentInProgress$.next(createEmptyStateResource());
  }

  private initUploadBescheidDocumentInProgress(fileName: string): void {
    this.uploadBescheidDocumentInProgress$.next({ fileName, loading: true });
  }

  public getUploadBescheidDocumentInProgress(): Observable<UploadFileInProgress> {
    return this.uploadBescheidDocumentInProgress$.asObservable();
  }

  doUploadBescheidDocument(bescheid: BescheidResource, file: File): void {
    this.binaryFileService
      .uploadFile(bescheid, BescheidLinkRel.UPLOAD_BESCHEID_FILE, file, false)
      .pipe(filterIsLoadedOrHasError(), first())
      .subscribe((binaryFileStateResource: StateResource<BinaryFileResource>) =>
        this.handleUploadBescheidDocumentResponse(bescheid, binaryFileStateResource),
      );
  }

  handleUploadBescheidDocumentResponse(
    bescheid: BescheidResource,
    binaryFileStateResource: StateResource<BinaryFileResource>,
  ): void {
    if (hasStateResourceError(binaryFileStateResource)) {
      this.setUploadBescheidDocumentInProgressError(binaryFileStateResource.error);
    } else {
      this.createBescheidDocumentFromFile(bescheid, binaryFileStateResource.resource);
    }
  }

  createBescheidDocumentFromFile(bescheid: BescheidResource, binaryFile: BinaryFileResource): void {
    this.commandService
      .createCommandByProps(buildCreateBescheidDocumentFromFileProps(bescheid, binaryFile))
      .pipe(filterIsLoadedOrHasError(), first())
      .subscribe((commandStateResource: StateResource<CommandResource>) =>
        this.handleCreateBescheidDocumentFromFileResponse(commandStateResource, binaryFile),
      );
  }

  handleCreateBescheidDocumentFromFileResponse(
    commandStateResource: StateResource<CommandResource>,
    binaryFile: BinaryFileResource,
  ): void {
    if (hasStateResourceError(commandStateResource)) {
      this.setUploadBescheidDocumentInProgressError(commandStateResource.error);
    } else {
      this.bescheidDocument$.next(createEmptyStateResource());
      this.bescheidDocumentFile$.next(createStateResource(binaryFile));
      this.bescheidDocumentUri$.next(getEffectedResourceUrl(commandStateResource.resource));
      this.clearUploadBescheidDocumentInProgress();
    }
  }

  private setUploadBescheidDocumentInProgressError(error: HttpError): void {
    this.uploadBescheidDocumentInProgress$.next({ loading: false, error });
  }

  public deleteBescheidDocument(): void {
    this.bescheidDocumentUri$.next(null);
    this.bescheidDocumentFile$.next(createEmptyStateResource());
    this.bescheidDocument$.next(createEmptyStateResource());
  }

  public createBescheidDocument(): Observable<StateResource<CommandResource>> {
    this.clearUploadBescheidDocumentInProgress();
    this.setCreateBescheidDocumentInProgress();
    this.doCreateBescheidDocument()
      .pipe(filterIsLoadedOrHasError(), first())
      .subscribe((commandStateResource: StateResource<CommandResource>) =>
        this.handleCreateBescheidDocumentResponse(commandStateResource),
      );
    return this.getCreateBescheidDocumentInProgress();
  }

  private clearUploadBescheidDocumentInProgress(): void {
    this.uploadBescheidDocumentInProgress$.next({ loading: false });
  }

  public getCreateBescheidDocumentInProgress(): Observable<StateResource<CommandResource>> {
    return this.createBescheidDocumentInProgress$.asObservable();
  }

  private setCreateBescheidDocumentInProgress(): void {
    this.createBescheidDocumentInProgress$.next(createEmptyStateResource(true));
  }

  doCreateBescheidDocument(): Observable<StateResource<CommandResource>> {
    return this.commandService.createCommandByProps(
      buildCreateBescheidDocumentCommandProps(this.getResource()),
    );
  }

  private handleCreateBescheidDocumentResponse(
    commandStateResource: StateResource<CommandResource>,
  ): void {
    this.createBescheidDocumentInProgress$.next(commandStateResource);
    if (notHasCommandError(commandStateResource.resource)) {
      const documentUri: ResourceUri = getEffectedResourceUrl(commandStateResource.resource);
      this.bescheidDocumentUri$.next(documentUri);
      this.loadBescheidDocumentByUri(documentUri);
    }
  }

  loadBescheidDocumentFile(document: DocumentResource): void {
    this.binaryFileService
      .getFile(getUrl(document, DocumentLinkRel.FILE))
      .pipe(filterIsLoadedOrHasError(), first())
      .subscribe((binaryFile) => {
        this.bescheidDocumentFile$.next(binaryFile);
        this.createBescheidDocumentInProgress$.next(createEmptyStateResource());
      });
  }

  public getBescheidDocumentCommand(): Observable<StateResource<CommandResource>> {
    return this.commandService.getCommandByOrder(CommandOrder.CREATE_BESCHEID_DOCUMENT);
  }

  public existsBescheidDraft(): boolean {
    let exists: boolean;
    this.bescheidResourceService
      .existResource()
      .pipe(take(1))
      .subscribe((existsDraft: boolean) => (exists = existsDraft));
    return exists;
  }

  public bescheidVerwerfen(): Observable<StateResource<CommandResource>> {
    return this.deleteBescheid(this.getResource()).pipe(
      tapOnCommandSuccessfullyDone(() => {
        this.deleteBescheidDocument();
        this.vorgangService.reloadCurrentVorgang();
      }),
    );
  }

  deleteBescheid(bescheid: BescheidResource): Observable<StateResource<CommandResource>> {
    return this.commandService.createCommandByProps(buildDeleteBescheidCommandProps(bescheid));
  }

  /**
   * @returns @deprecated Don't use this function, instead pass data to function if necessarry.
   */
  getResource(): BescheidResource {
    let resource: StateResource<BescheidResource> = undefined;
    this.bescheidResourceService
      .selectResource()
      .pipe(take(1))
      .subscribe((stateResource: StateResource<BescheidResource>) => (resource = stateResource));

    return resource.resource;
  }

  public reloadCurrentVorgang(): void {
    this.vorgangService.reloadCurrentVorgang();
  }

  public loadBescheidDocument(
    resourceUri: ResourceUri,
  ): Observable<StateResource<DocumentResource>> {
    return this.repository.getResource<DocumentResource>(resourceUri).pipe(
      map((documentResource: DocumentResource) => createStateResource(documentResource)),
      startWith(createEmptyStateResource<DocumentResource>(true)),
    );
  }

  public getEmpfaenger(): Observable<string> {
    return this.vorgangService.getVorgangWithEingang().pipe(
      filter(isLoaded),
      map((stateResource) => stateResource.resource),
      map(getEmpfaenger),
    );
  }

  public getLastBescheid(): Observable<BescheidResource> {
    return this.bescheidListResourceService.getItems().pipe(
      map((bescheide: BescheidResource[]) => this.filterBySentStatus(bescheide)),
      map((bescheide: BescheidResource[]) => this.sortByBeschiedenAm(bescheide)),
      map((bescheide: BescheidResource[]) => bescheide[0]),
    );
  }

  private sortByBeschiedenAm(bescheide: BescheidResource[]): BescheidResource[] {
    return sortByGermanDateStr<BescheidResource>(
      bescheide,
      (bescheid: BescheidResource) => bescheid.beschiedenAm,
    );
  }

  public existBescheid(): Observable<boolean> {
    return this.bescheidListResourceService.getItems().pipe(
      map((bescheide: BescheidResource[]) => this.filterBySentStatus(bescheide)),
      map((bescheide: BescheidResource[]) => isNotEmpty(bescheide)),
    );
  }

  public getBescheidList(): Observable<StateResource<BescheidListResource>> {
    return this.bescheidListResourceService.getList();
  }

  filterBySentStatus(bescheide: BescheidResource[]): BescheidResource[] {
    return isNotNil(bescheide) ? bescheide.filter(this.hasSentStatus) : EMPTY_ARRAY;
  }

  private hasSentStatus(bescheid: BescheidResource): boolean {
    return bescheid.status === BescheidStatus.SENT;
  }

  public refreshList(): void {
    this.bescheidListResourceService.refresh();
  }

  public uploadAttachment(
    bescheidResource: BescheidResource,
    file: File,
  ): Observable<StateResource<BinaryFileResource>> {
    this.uploadAttachmentInProgress$.next({ fileName: file.name, loading: true });
    return this.binaryFileService
      .uploadFile(bescheidResource, BescheidLinkRel.UPLOAD_ATTACHMENT, file, false)
      .pipe(
        tap((binaryFileStateResource: StateResource<BinaryFileResource>) =>
          this.handleAttachmentUpload(binaryFileStateResource),
        ),
      );
  }

  handleAttachmentUpload(binaryFileStateResource: StateResource<BinaryFileResource>) {
    if (hasStateResourceError(binaryFileStateResource)) {
      this.uploadAttachmentInProgress$.next({
        loading: false,
        error: binaryFileStateResource.error,
      });
    } else {
      this.uploadAttachmentInProgress$.next({
        ...this.uploadAttachmentInProgress$.value,
        loading: binaryFileStateResource.loading,
      });
    }
    this.uploadedAttachment$.next(binaryFileStateResource);
  }

  public getUploadAttachmentInProgress(): Observable<UploadFileInProgress> {
    return this.uploadAttachmentInProgress$.asObservable();
  }

  public getUploadedAttachment(): Observable<StateResource<BinaryFileResource>> {
    return this.uploadedAttachment$.asObservable();
  }

  public clearAttachmentUpload(): void {
    this.uploadAttachmentInProgress$.next(createEmptyStateResource());
  }
}
