import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import { CommandOrder, CreateCommand, CreateCommandProps } from '@alfa-client/command-shared';
import { EMPTY_STRING, StateResource, createStateResource } from '@alfa-client/tech-shared';
import { getUrl } from '@ngxp/rest';
import { createBinaryFileResource } from 'libs/binary-file-shared/test/binary-file';
import { createApiError } from '../../../tech-shared/test/error';
import {
  createBescheid,
  createBescheidListResource,
  createBescheidResource,
} from '../test/bescheid';
import { BescheidLinkRel, BescheidListLinkRel } from './bescheid.linkrel';
import { Bescheid, BescheidListResource, BescheidResource } from './bescheid.model';
import {
  buildCreateBescheidCommand,
  buildCreateBescheidDocumentCommandProps,
  buildCreateBescheidDocumentFromFileProps,
  buildDeleteBescheidCommandProps,
  buildSendBescheidCommandProps,
  buildUpdateBescheidCommandProps,
  getEmbeddedBescheidResources,
  hasUploadNoError,
  isCreateBescheidCommand,
  isUploadFinished,
} from './bescheid.util';

import * as ResourceUtil from '../../../tech-shared/src/lib/resource/resource.util';

describe('BescheidUtil', () => {
  describe('buildCreateBescheidCommand', () => {
    it('should have order', () => {
      const command: CreateCommand = buildCreateBescheidCommand();

      expect(command.order).toEqual(CommandOrder.CREATE_BESCHEID);
    });

    it('should have null as body', () => {
      const command: CreateCommand = buildCreateBescheidCommand();

      expect(command.body).toBeNull();
    });

    it('should have body', () => {
      const bescheid = createBescheid();

      const command = buildCreateBescheidCommand(bescheid);

      expect(command.body).toEqual(bescheid);
    });
  });

  describe('isCreateBescheidCommand', () => {
    it('should return true on matching command', () => {
      const isCommand: boolean = isCreateBescheidCommand(CommandOrder.CREATE_BESCHEID);

      expect(isCommand).toBeTruthy();
    });
  });

  describe('build create bescheid document from file props', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const binaryFileResource: BinaryFileResource = createBinaryFileResource();

    it('should have resource', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentFromFileProps(
        bescheid,
        binaryFileResource,
      );

      expect(props.resource).toBe(bescheid);
    });

    it('should have linkRel', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentFromFileProps(
        bescheid,
        binaryFileResource,
      );

      expect(props.linkRel).toBe(BescheidLinkRel.CREATE_DOCUMENT_FROM_FILE);
    });

    describe('command', () => {
      it('should have order', () => {
        const props: CreateCommandProps = buildCreateBescheidDocumentFromFileProps(
          bescheid,
          binaryFileResource,
        );

        expect(props.command.order).toBe(CommandOrder.CREATE_BESCHEID_DOCUMENT_FROM_FILE);
      });

      it('should have documentFile in body', () => {
        const props: CreateCommandProps = buildCreateBescheidDocumentFromFileProps(
          bescheid,
          binaryFileResource,
        );

        expect(props.command.body.documentFile).toBe(getUrl(binaryFileResource));
      });
    });

    it('should have snackBarMessage', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentFromFileProps(
        bescheid,
        binaryFileResource,
      );

      expect(props.snackBarMessage).toBe(EMPTY_STRING);
    });
  });

  describe(' build update bescheid command props', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const bescheid: Bescheid = createBescheid();

    it('should have resource', () => {
      const props: CreateCommandProps = buildUpdateBescheidCommandProps(bescheidResource, bescheid);

      expect(props.resource).toBe(bescheidResource);
    });

    it('should have linkRel', () => {
      const props: CreateCommandProps = buildUpdateBescheidCommandProps(bescheidResource, bescheid);

      expect(props.linkRel).toBe(BescheidLinkRel.UPDATE);
    });

    describe('command', () => {
      it('should have order', () => {
        const props: CreateCommandProps = buildUpdateBescheidCommandProps(
          bescheidResource,
          bescheid,
        );

        expect(props.command.order).toBe(CommandOrder.UPDATE_BESCHEID);
      });

      describe('body', () => {
        it('should have bescheid', () => {
          const props: CreateCommandProps = buildUpdateBescheidCommandProps(
            bescheidResource,
            bescheid,
          );

          expect(props.command.body).toBe(bescheid);
        });
      });

      it('should have snackBarMessage', () => {
        const props: CreateCommandProps = buildUpdateBescheidCommandProps(
          bescheidResource,
          bescheid,
        );

        expect(props.snackBarMessage).toBe(EMPTY_STRING);
      });
    });
  });

  describe('build delete bescheid command props', () => {
    const bescheidResource: BescheidResource = createBescheidResource();

    it('should have resource', () => {
      const props: CreateCommandProps = buildDeleteBescheidCommandProps(bescheidResource);

      expect(props.resource).toBe(bescheidResource);
    });

    it('should have linkRel', () => {
      const props: CreateCommandProps = buildDeleteBescheidCommandProps(bescheidResource);

      expect(props.linkRel).toBe(BescheidLinkRel.DELETE);
    });

    describe('command', () => {
      it('should have order', () => {
        const props: CreateCommandProps = buildDeleteBescheidCommandProps(bescheidResource);

        expect(props.command.order).toBe(CommandOrder.DELETE_BESCHEID);
      });

      it('should have body', () => {
        const props: CreateCommandProps = buildDeleteBescheidCommandProps(bescheidResource);

        expect(props.command.body).toBeNull();
      });
    });

    it('should have snackBarMessage', () => {
      const props: CreateCommandProps = buildDeleteBescheidCommandProps(bescheidResource);

      expect(props.snackBarMessage).toBe(EMPTY_STRING);
    });
  });

  describe('build create bescheid document command props', () => {
    const bescheidResource: BescheidResource = createBescheidResource();

    it('should have resource', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentCommandProps(bescheidResource);

      expect(props.resource).toBe(bescheidResource);
    });

    it('should have linkRel', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentCommandProps(bescheidResource);

      expect(props.linkRel).toBe(BescheidLinkRel.CREATE_DOCUMENT);
    });

    describe('command', () => {
      it('should have order', () => {
        const props: CreateCommandProps = buildCreateBescheidDocumentCommandProps(bescheidResource);

        expect(props.command.order).toBe(CommandOrder.CREATE_BESCHEID_DOCUMENT);
      });

      it('should have body', () => {
        const props: CreateCommandProps = buildCreateBescheidDocumentCommandProps(bescheidResource);

        expect(props.command.body).toBeNull();
      });
    });

    it('should have snackBarMessage', () => {
      const props: CreateCommandProps = buildCreateBescheidDocumentCommandProps(bescheidResource);

      expect(props.snackBarMessage).toBe(EMPTY_STRING);
    });
  });

  describe('build send bescheid command props', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const linkRel: string = 'link_rel';

    it('should have resource', () => {
      const props: CreateCommandProps = buildSendBescheidCommandProps(bescheidResource, linkRel);

      expect(props.resource).toBe(bescheidResource);
    });

    it('should have linkRel', () => {
      const props: CreateCommandProps = buildSendBescheidCommandProps(bescheidResource, linkRel);

      expect(props.linkRel).toBe(linkRel);
    });

    describe('command', () => {
      it('should have order', () => {
        const props: CreateCommandProps = buildSendBescheidCommandProps(bescheidResource, linkRel);

        expect(props.command.order).toBe(CommandOrder.SEND_BESCHEID);
      });

      it('should have body', () => {
        const props: CreateCommandProps = buildSendBescheidCommandProps(bescheidResource, linkRel);

        expect(props.command.body).toBeNull();
      });
    });

    it('should have snackBarMessage', () => {
      const props: CreateCommandProps = buildSendBescheidCommandProps(bescheidResource, linkRel);

      expect(props.snackBarMessage).toBe(EMPTY_STRING);
    });
  });

  describe('get embedded bescheid resources', () => {
    const bescheidListStateResource: StateResource<BescheidListResource> = createStateResource(
      createBescheidListResource(),
    );
    const embeddedBescheidResources: BescheidResource[] = [createBescheidResource()];
    let getEmbeddedResourcesSpy: jest.SpyInstance;

    beforeEach(() => {
      getEmbeddedResourcesSpy = jest
        .spyOn(ResourceUtil, 'getEmbeddedResources')
        .mockReturnValue(embeddedBescheidResources);
    });

    it('should call get embedded resources', () => {
      getEmbeddedBescheidResources(bescheidListStateResource);

      expect(getEmbeddedResourcesSpy).toHaveBeenCalledWith(
        bescheidListStateResource,
        BescheidListLinkRel.BESCHEID_LIST,
      );
    });

    it('should return embedded resources', () => {
      const embeddedResource: BescheidResource[] =
        getEmbeddedBescheidResources(bescheidListStateResource);

      expect(embeddedResource).toBe(embeddedBescheidResources);
    });
  });

  describe('isUploadFinished', () => {
    it('should return true', () => {
      const isFinished: boolean = isUploadFinished({ loading: false });

      expect(isFinished).toBeTruthy();
    });

    it('should return false', () => {
      const isFinished: boolean = isUploadFinished({ loading: true });

      expect(isFinished).toBeFalsy();
    });
  });

  describe('hasUploadNoError', () => {
    it('should return true', () => {
      const hasNoError: boolean = hasUploadNoError({ error: null, loading: false });

      expect(hasNoError).toBeTruthy();
    });

    it('should return false', () => {
      const hasNoError: boolean = hasUploadNoError({ error: createApiError(), loading: false });

      expect(hasNoError).toBeFalsy();
    });
  });
});
