import { CollaborationListResource } from '@alfa-client/collaboration-shared';
import { CommandLinkRel, CommandResource } from '@alfa-client/command-shared';
import { StateResource, createStateResource } from '@alfa-client/tech-shared';
import { dispatchEventFromFixture, getMockComponent, mock, useFromMock } from '@alfa-client/test-utils';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { FormBuilder, ReactiveFormsModule } from '@angular/forms';
import { ButtonWithSpinnerComponent, TextEditorComponent, TextareaEditorComponent } from '@ods/component';
import { ButtonComponent, CloseIconComponent } from '@ods/system';
import { CollaborationService } from 'libs/collaboration-shared/src/lib/collaboration.service';
import { createCollaborationListResource } from 'libs/collaboration-shared/test/collaboration';
import { createCommandResource } from 'libs/command-shared/test/command';
import { getDataTestIdAttributeOf } from 'libs/tech-shared/test/data-test';
import { singleColdCompleted } from 'libs/tech-shared/test/marbles';
import { MockComponent } from 'ng-mocks';
import { of } from 'rxjs';
import { CollaborationRequestFormComponent } from './collaboration-request-form.component';
import { CollaborationRequestFormService } from './collaboration.request.formservice';

describe('CollaborationRequestFormComponent', () => {
  let component: CollaborationRequestFormComponent;
  let fixture: ComponentFixture<CollaborationRequestFormComponent>;

  const cancelButton: string = getDataTestIdAttributeOf('collaboration-request-cancel-button');
  const submitButton: string = getDataTestIdAttributeOf('collaboration-request-submit-button');

  let formService: CollaborationRequestFormService;

  const stateCommandResource: StateResource<CommandResource> = createStateResource(createCommandResource());

  beforeEach(async () => {
    formService = new CollaborationRequestFormService(new FormBuilder(), useFromMock(mock(CollaborationService)));
    TestBed.overrideComponent(CollaborationRequestFormComponent, {
      set: {
        providers: [
          {
            provide: CollaborationRequestFormService,
            useValue: formService,
          },
        ],
      },
    });
    await TestBed.configureTestingModule({
      imports: [ReactiveFormsModule],
      declarations: [
        CollaborationRequestFormComponent,
        MockComponent(ButtonComponent),
        MockComponent(ButtonWithSpinnerComponent),
        MockComponent(CloseIconComponent),
        MockComponent(TextEditorComponent),
        MockComponent(TextareaEditorComponent),
      ],
      providers: [
        {
          provide: CollaborationRequestFormService,
          useValue: formService,
        },
      ],
    }).compileComponents();
    fixture = TestBed.createComponent(CollaborationRequestFormComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('cancel button', () => {
    it('should emit hideRequestForm', () => {
      component.abbrechen.emit = jest.fn();

      dispatchEventFromFixture(fixture, cancelButton, 'clickEmitter');

      expect(component.abbrechen.emit).toHaveBeenCalled();
    });
  });

  describe('submit button', () => {
    describe('component', () => {
      it('should be called with state resource', () => {
        component.submitInProgress$ = of(stateCommandResource);

        fixture.detectChanges();

        const submitButtonComp: ButtonWithSpinnerComponent = getMockComponent<ButtonWithSpinnerComponent>(
          fixture,
          ButtonWithSpinnerComponent,
        );
        expect(submitButtonComp.stateResource).toBe(stateCommandResource);
      });
    });

    it('should call submit on output', () => {
      component.submit = jest.fn();

      dispatchEventFromFixture(fixture, submitButton, 'clickEmitter');

      expect(component.submit).toHaveBeenCalled();
    });
  });

  describe('submit', () => {
    it('should call doSubmit', () => {
      component.doSubmit = jest.fn();

      component.submit();

      expect(component.doSubmit).toHaveBeenCalled();
    });

    it('should set submitInProgress', () => {
      component.doSubmit = jest.fn().mockReturnValue(of(stateCommandResource));

      component.submit();

      expect(component.submitInProgress$).toBeObservable(singleColdCompleted(stateCommandResource));
    });
  });

  describe('do submit', () => {
    beforeEach(() => {
      formService.submit = jest
        .fn()
        .mockReturnValue(of(createStateResource(createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]))));
    });

    it('should call formService', () => {
      formService.submit = jest.fn().mockReturnValue(of(stateCommandResource));

      component.doSubmit().subscribe();

      expect(formService.submit).toHaveBeenCalled();
    });

    it('should emit show result', () => {
      component.showResult.emit = jest.fn();

      component.doSubmit().subscribe();

      expect(component.showResult.emit).toHaveBeenCalledWith(formService.form.value);
    });

    it('should return value', () => {
      formService.submit = jest.fn().mockReturnValue(of(stateCommandResource));

      expect(component.doSubmit()).toBeObservable(singleColdCompleted(stateCommandResource));
    });
  });

  describe('set collaboration list resource', () => {
    it('should call set list resource on formService', () => {
      formService.setListResource = jest.fn();
      const collaborationListResource: CollaborationListResource = createCollaborationListResource();

      component.collaborationListResource = collaborationListResource;

      expect(formService.setListResource).toHaveBeenCalledWith(collaborationListResource);
    });
  });
});
