import { EMPTY_ARRAY, StateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { faker } from '@faker-js/faker';
import { InstantSearchResult } from '@ods/system';
import { createInstantSearchResult } from 'libs/design-system/src/test/search';
import { createEmptyListResource } from 'libs/tech-shared/test/resource';
import { Observable, of } from 'rxjs';
import { singleColdCompleted } from '../../../../tech-shared/test/marbles';
import { createExterneFachstelleListResource, createExterneFachstelleResource } from './../../../test/externe-fachstelle';
import { ExterneFachstelleResourceSearchService } from './externe-fachstelle-resource-search.service';
import { ExterneFachstelleListResource, ExterneFachstelleResource } from './externe-fachstelle.model';
import { ExterneFachstelleService } from './externe-fachstelle.service';

jest.mock('./externe-fachstelle-resource-search.service');

describe('ExterneFachstelleService', () => {
  let service: ExterneFachstelleService;

  let searchService: Mock<ExterneFachstelleResourceSearchService>;

  const listResource: ExterneFachstelleListResource = createExterneFachstelleListResource();
  const listStateResource: StateResource<ExterneFachstelleListResource> = createStateResource(listResource);

  beforeEach(() => {
    searchService = mock(ExterneFachstelleResourceSearchService);

    service = new ExterneFachstelleService(useFromMock(searchService));
  });

  describe('get search result list', () => {
    const instantSearchResult: InstantSearchResult<ExterneFachstelleResource> = createInstantSearchResult();

    beforeEach(() => {
      searchService.getResultList.mockReturnValue(of(listStateResource));
      service.mapListResource = jest.fn().mockReturnValue([instantSearchResult]);
    });

    it('should call search service', () => {
      service.getSearchResultList();

      expect(searchService.getResultList).toHaveBeenCalled();
    });

    it('should map list resource', () => {
      service.getSearchResultList().subscribe();

      expect(service.mapListResource).toHaveBeenCalledWith(listStateResource);
    });

    it('should return result', (done) => {
      searchService.getResultList.mockReturnValue(of(listStateResource));

      service.getSearchResultList().subscribe((result: InstantSearchResult<ExterneFachstelleResource>[]) => {
        expect(result).toEqual([instantSearchResult]);
        done();
      });
    });
  });

  describe('map listResource', () => {
    const externeFachstelleResource: ExterneFachstelleResource = createExterneFachstelleResource();
    const singleEntrylistResource: ExterneFachstelleListResource = createExterneFachstelleListResource([
      externeFachstelleResource,
    ]);
    const singleEntryListStateResource: StateResource<ExterneFachstelleListResource> =
      createStateResource(singleEntrylistResource);

    const instantSearchResult: InstantSearchResult<ExterneFachstelleResource> = createInstantSearchResult();

    beforeEach(() => {
      service.mapToInstantSearchResult = jest.fn().mockReturnValue(instantSearchResult);
    });

    it('should map to instantSearchResult', () => {
      service.mapListResource(singleEntryListStateResource);

      expect(service.mapToInstantSearchResult).toHaveBeenCalledWith(externeFachstelleResource);
    });

    it('should return value', () => {
      const result: InstantSearchResult<ExterneFachstelleResource>[] = service.mapListResource(singleEntryListStateResource);

      expect(result).toEqual([instantSearchResult]);
    });

    it('should return empty array on empty listResource', () => {
      const result: InstantSearchResult<ExterneFachstelleResource>[] = service.mapListResource(
        createStateResource(createEmptyListResource<ExterneFachstelleListResource>()),
      );

      expect(result).toEqual(EMPTY_ARRAY);
    });
  });

  describe('map to instant search result', () => {
    const externeFachstelleResource: ExterneFachstelleResource = createExterneFachstelleResource();

    it('should map titel', () => {
      const instantSearchResult: InstantSearchResult<ExterneFachstelleResource> =
        service.mapToInstantSearchResult(externeFachstelleResource);

      expect(instantSearchResult.title).toBe(externeFachstelleResource.name);
    });
    it('should map description', () => {
      const instantSearchResult: InstantSearchResult<ExterneFachstelleResource> =
        service.mapToInstantSearchResult(externeFachstelleResource);

      expect(instantSearchResult.description).toBe(externeFachstelleResource.anschrift);
    });

    it('should map data', () => {
      const instantSearchResult: InstantSearchResult<ExterneFachstelleResource> =
        service.mapToInstantSearchResult(externeFachstelleResource);

      expect(instantSearchResult.data).toBe(externeFachstelleResource);
    });
  });

  describe('search', () => {
    const searchBy: string = faker.lorem.word();

    it('should call search service with search string', () => {
      service.search(searchBy);

      expect(searchService.search).toHaveBeenCalledWith(searchBy);
    });
  });

  describe('clear search result', () => {
    it('should call search service', () => {
      service.clearSearchResult();

      expect(searchService.clearResultList).toHaveBeenCalledWith();
    });
  });

  describe('get selected result', () => {
    const externeFachstelleResource: ExterneFachstelleResource = createExterneFachstelleResource();

    beforeEach(() => {
      searchService.getSelectedResult.mockReturnValue(of(externeFachstelleResource));
    });

    it('should call service', () => {
      service.getSelectedResult();

      expect(searchService.getSelectedResult).toHaveBeenCalled();
    });

    it('should return result', () => {
      const selectedResult$: Observable<ExterneFachstelleResource> = service.getSelectedResult();

      expect(selectedResult$).toBeObservable(singleColdCompleted(externeFachstelleResource));
    });
  });

  describe('select search result', () => {
    const externeFachstelleResource: ExterneFachstelleResource = createExterneFachstelleResource();

    it('should call service', () => {
      service.selectSearchResult(externeFachstelleResource);

      expect(searchService.selectResult).toHaveBeenCalledWith(externeFachstelleResource);
    });
  });

  describe('clear selected result', () => {
    it('should call service', () => {
      service.clearSelectedResult();

      expect(searchService.clearSelectedResult).toHaveBeenCalled();
    });
  });
});
