import { ToEmbeddedResourcesPipe } from '@alfa-client/tech-shared';
import { EventData, Mock, dialogRefMock, getMockComponent, mock, triggerEvent } from '@alfa-client/test-utils';
import {
  OrganisationsEinheitResource,
  OrganisationsEinheitService,
  ZUSTAENDIGE_STELLE_SERVICE,
} from '@alfa-client/zustaendige-stelle-shared';
import { DIALOG_DATA, DialogRef } from '@angular/cdk/dialog';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { faker } from '@faker-js/faker';
import { Resource } from '@ngxp/rest';
import { ButtonComponent, CloseIconComponent, InstantSearchResult } from '@ods/system';
import { createInstantSearchResult } from 'libs/design-system/src/test/search';
import { MockComponent } from 'ng-mocks';
import { of } from 'rxjs';
import { getDataTestIdOf } from '../../../../tech-shared/test/data-test';
import { createOrganisationsEinheitResource } from '../../../../zustaendige-stelle-shared/test/organisations-einheit';
import { SearchZustaendigeStelleDialogComponent } from '././search-zustaendige-stelle-dialog.component';
import { SearchZustaendigeStelleFormComponent } from './search-zustaendige-stelle-form/search-zustaendige-stelle-form.component';

describe('SearchZustaendigeStelleDialogComponent', () => {
  let component: SearchZustaendigeStelleDialogComponent<Resource>;
  let fixture: ComponentFixture<SearchZustaendigeStelleDialogComponent<Resource>>;

  const searchOrganisationsEinheitComp: string = getDataTestIdOf('search-organisations-einheit');

  const service: Mock<OrganisationsEinheitService> = mock(OrganisationsEinheitService);
  const title: string = 'dummyTitle';

  const organisationsEinheitResource: OrganisationsEinheitResource = createOrganisationsEinheitResource();

  beforeEach(async () => {
    await TestBed.configureTestingModule({
      declarations: [
        SearchZustaendigeStelleDialogComponent,
        ToEmbeddedResourcesPipe,
        MockComponent(SearchZustaendigeStelleFormComponent),
        MockComponent(ButtonComponent),
        MockComponent(CloseIconComponent),
      ],
      providers: [
        {
          provide: ZUSTAENDIGE_STELLE_SERVICE,
          useValue: service,
        },
        {
          provide: DialogRef,
          useValue: dialogRefMock,
        },
        {
          provide: DIALOG_DATA,
          useValue: title,
        },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(SearchZustaendigeStelleDialogComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('ngOnInit', () => {
    it('should call service', () => {
      component.ngOnInit();

      expect(service.getSearchResultList).toHaveBeenCalled();
    });
  });

  describe('onKeyDownHandler', () => {
    it('should prevent default behavior for enter key on non buttons', () => {
      const keyboardEvent: KeyboardEvent = {
        ...new KeyboardEvent('enter'),
        key: 'Enter',
        preventDefault: jest.fn(),
        target: new EventTarget(),
      };

      component.onKeyDownHandler(keyboardEvent);

      expect(keyboardEvent.preventDefault).toHaveBeenCalled();
    });

    it('should not prevent default behavior for enter key on buttons', () => {
      const keyboardEvent: KeyboardEvent = {
        ...new KeyboardEvent('enter'),
        key: 'Enter',
        preventDefault: jest.fn(),
        target: { ...new EventTarget(), tagName: 'BUTTON' } as Element,
      };

      component.onKeyDownHandler(keyboardEvent);

      expect(keyboardEvent.preventDefault).not.toHaveBeenCalled();
    });
  });

  describe('search organisationsEinheit component', () => {
    const result: InstantSearchResult<Resource> = createInstantSearchResult();

    beforeEach(() => {
      component.searchResults$ = of([result]);
      fixture.detectChanges();
    });

    it('should be called with searchResult', () => {
      const comp: SearchZustaendigeStelleFormComponent<Resource> = getMockComponent<
        SearchZustaendigeStelleFormComponent<Resource>
      >(fixture, SearchZustaendigeStelleFormComponent);

      expect(comp.searchResults).toEqual([result]);
    });

    it('should call search on openSearchDialog output', () => {
      component.search = jest.fn();
      const searchBy: string = faker.word.sample();
      const eventData: EventData<SearchZustaendigeStelleDialogComponent<Resource>> = {
        fixture,
        elementSelector: searchOrganisationsEinheitComp,
        name: 'search',
        data: searchBy,
      };

      triggerEvent(eventData);

      expect(component.search).toHaveBeenCalledWith(searchBy);
    });

    it('should call selectSearchResult on selectSearchResult output', () => {
      component.selectSearchResult = jest.fn();

      const eventData: EventData<SearchZustaendigeStelleDialogComponent<Resource>> = {
        fixture,
        elementSelector: searchOrganisationsEinheitComp,
        name: 'selectSearchResult',
        data: organisationsEinheitResource,
      };

      triggerEvent(eventData);

      expect(component.selectSearchResult).toHaveBeenCalledWith(organisationsEinheitResource);
    });

    it('should call clearSearchResult', () => {
      component.clearSearchResult = jest.fn();

      const eventData: EventData<SearchZustaendigeStelleDialogComponent<Resource>> = {
        fixture,
        elementSelector: searchOrganisationsEinheitComp,
        name: 'clearSearchResult',
        data: organisationsEinheitResource,
      };

      triggerEvent(eventData);

      expect(component.clearSearchResult).toHaveBeenCalled();
    });
  });

  describe('search', () => {
    const searchBy: string = faker.word.sample();

    it('should call service', () => {
      component.search(searchBy);

      expect(service.search).toHaveBeenCalledWith(searchBy);
    });
  });

  describe('select search result', () => {
    it('should set select result', () => {
      component.selectSearchResult(organisationsEinheitResource);

      expect(service.selectSearchResult).toHaveBeenCalledWith(organisationsEinheitResource);
    });

    it('should call service', () => {
      component.selectSearchResult(organisationsEinheitResource);

      expect(service.clearSearchResult).toHaveBeenCalled();
    });

    it('should close dialog', () => {
      component.selectSearchResult(organisationsEinheitResource);

      expect(dialogRefMock.close).toHaveBeenCalled();
    });
  });

  describe('clear search result', () => {
    it('should call service', () => {
      component.clearSearchResult();

      expect(service.clearSearchResult).toHaveBeenCalled();
    });
  });

  describe('close dialog', () => {
    it('should clear search result', () => {
      component.clearSearchResult = jest.fn();

      component.closeDialog();

      expect(component.clearSearchResult).toHaveBeenCalled();
    });

    it('should close dialog', () => {
      component.closeDialog();

      expect(dialogRefMock.close).toHaveBeenCalled();
    });
  });

  describe('isNotButtonElement', () => {
    it('should return false', () => {
      const element: Element = { tagName: 'BUTTON' } as Element;

      const result: boolean = component.isNotButtonElement(element);

      expect(result).toBe(false);
    });

    it('should return true', () => {
      const element: Element = { tagName: 'INPUT' } as Element;

      const result: boolean = component.isNotButtonElement(element);

      expect(result).toBe(true);
    });
  });
});
