package de.ozgcloud.alfa.collaboration;

import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.Collections;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.util.UriComponentsBuilder;

import lombok.SneakyThrows;

class CollaborationControllerTest {

	@InjectMocks
	private CollaborationController controller;

	@Mock
	private CollaborationService service;

	@Mock
	private CollaborationModelAssembler assembler;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestGetAllByVorgangId {

		private final Collaboration collaboration = CollaborationTestFactory.create();

		@SneakyThrows
		@Test
		void shouldCallCollaborationService() {
			performRequest();

			verify(service).getCollaborations(CollaborationTestFactory.VORGANG_ID);
		}

		@Nested
		class TestOnExistingCollaboration {

			private final Stream<Collaboration> collaborations = Stream.of(collaboration);

			@BeforeEach
			void mock() {
				when(service.getCollaborations(CollaborationTestFactory.VORGANG_ID)).thenReturn(collaborations);
			}

			@Test
			void shouldCallAssembler() {
				performRequest();

				verify(assembler).toCollectionModel(collaborations, CollaborationTestFactory.VORGANG_ID);
			}

			@SneakyThrows
			@Test
			void shouldReturnStatusOk() {
				when(assembler.toCollectionModel(collaborations, CollaborationTestFactory.VORGANG_ID))
						.thenReturn(CollectionModel.of(Collections.singletonList(EntityModel.of(collaboration))));

				var response = performRequest();

				response.andExpect(status().isOk());
			}

			@Nested
			class TestResponseBody {

				@BeforeEach
				void mockAssembler() {
					when(assembler.toCollectionModel(collaborations, CollaborationTestFactory.VORGANG_ID))
							.thenReturn(CollectionModel.of(Collections.singletonList(EntityModel.of(collaboration))));
				}

				@SneakyThrows
				@Test
				void shouldNotHaveVorgangId() {
					var response = performRequest();

					response.andExpect(jsonPath("$.vorgangId").doesNotExist());
				}

				@SneakyThrows
				@Test
				void shouldNotHaveCollaborationVorgangId() {
					var response = performRequest();

					response.andExpect(jsonPath("$.collaborationVorgangId").doesNotExist());
				}

				@SneakyThrows
				@Test
				void shouldHaveTitel() {
					var response = performRequest();

					response.andExpect(jsonPath("$.content[0].titel").value(CollaborationTestFactory.TITEL));
				}

				@SneakyThrows
				@Test
				void shouldHaveBeschreibung() {
					var response = performRequest();

					System.out.println(response.andReturn().getResponse().getContentAsString());

					response.andExpect(jsonPath("$.content[0].beschreibung").value(CollaborationTestFactory.BESCHREIBUNG));
				}

				@SneakyThrows
				@Test
				void shouldHaveZustaendigeStelle() {
					var expectedLink = UriComponentsBuilder
							.fromUriString("http://localhost")
							.path(OrganisationsEinheitController.PATH)
							.pathSegment(CollaborationTestFactory.ZUSTAENDIGE_STELLE)
							.build();

					var response = performRequest();

					response.andExpect(jsonPath("$.content[0].zustaendigeStelle")
							.value(expectedLink.toString()));
				}
			}
		}

		@SneakyThrows
		private ResultActions performRequest() {
			return mockMvc.perform(get(CollaborationController.PATH + "/" + CollaborationTestFactory.VORGANG_ID + "/collaborations"));
		}
	}
}
