package de.ozgcloud.alfa.collaboration;

import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import com.thedeanda.lorem.LoremIpsum;

import lombok.SneakyThrows;

class FachstelleControllerTest {

	@InjectMocks
	private FachstelleController controller;

	@Mock
	private FachstelleService service;

	@Mock
	private FachstelleModelAssembler modelAssembler;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestGetById {

		private final Fachstelle fachstelle = FachstelleTestFactory.create();

		@BeforeEach
		void mockService() {
			when(service.getById(FachstelleTestFactory.ID)).thenReturn(fachstelle);
		}

		@SneakyThrows
		@Test
		void shouldCallFachstelleService() {
			performRequest();

			verify(service).getById(FachstelleTestFactory.ID);
		}

		@Test
		void shouldCallAssembler() {
			performRequest();

			verify(modelAssembler).toModel(fachstelle);
		}

		@SneakyThrows
		@Test
		void shouldReturnStatusOk() {
			when(modelAssembler.toModel(fachstelle)).thenReturn(EntityModel.of(fachstelle));

			var response = performRequest();

			response.andExpect(status().isOk());
		}

		@SneakyThrows
		@Test
		void shouldHaveFachstelleName() {
			when(modelAssembler.toModel(fachstelle)).thenReturn(EntityModel.of(fachstelle));

			var response = performRequest();

			response.andExpect(jsonPath("$.name").value(FachstelleTestFactory.NAME));
		}

		@SneakyThrows
		@Test
		void shouldHaveFachstelleAnschrift() {
			when(modelAssembler.toModel(fachstelle)).thenReturn(EntityModel.of(fachstelle));

			var response = performRequest();

			response.andExpect(jsonPath("$.anschrift").value(FachstelleTestFactory.ANSCHRIFT));
		}

		@SneakyThrows
		@Test
		void shouldHaveFachstelleEmail() {
			when(modelAssembler.toModel(fachstelle)).thenReturn(EntityModel.of(fachstelle));

			var response = performRequest();

			response.andExpect(jsonPath("$.email").value(FachstelleTestFactory.EMAIL));
		}

		@SneakyThrows
		private ResultActions performRequest() {
			return mockMvc.perform(get(FachstelleController.PATH + "/" + FachstelleTestFactory.ID));
		}
	}

	@Nested
	class TestSearch {

		private final Fachstelle fachstelle1 = FachstelleTestFactory.create();
		private final Fachstelle fachstelle2 = FachstelleTestFactory.createBuilder()
				.name(LoremIpsum.getInstance().getName())
				.build();
		private final String searchBy = LoremIpsum.getInstance().getWords(5);
		private final List<Fachstelle> fachstellen = List.of(fachstelle1, fachstelle2);

		@BeforeEach
		void setUpMocks() {
			when(service.searchFachstellen(searchBy)).thenReturn(Stream.of(fachstelle1, fachstelle2));
			when(modelAssembler.toCollectionModel(fachstellen))
					.thenReturn(
							CollectionModel.of(List.of(EntityModel.of(fachstelle1), EntityModel.of(fachstelle2))));
		}

		@Test
		void shouldCallService() {
			performRequest();

			verify(service).searchFachstellen(searchBy);
		}

		@Test
		void shouldCallModelAssembler() {
			performRequest();

			verify(modelAssembler).toCollectionModel(fachstellen);
		}

		@SneakyThrows
		@Test
		void shouldReturnStatusOk() {
			var response = performRequest();

			response.andExpect(status().isOk());
		}

		@Test
		@SneakyThrows
		void shouldHaveFirstFachstelleName() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[0].name").value(fachstelle1.getName()));
		}

		@Test
		@SneakyThrows
		void shouldHaveFirstFachstelleAnschrift() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[0].anschrift").value(fachstelle1.getAnschrift()));
		}

		@Test
		@SneakyThrows
		void shouldHaveFirstFachstelleEmail() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[0].email").value(fachstelle1.getEmail()));
		}

		@Test
		@SneakyThrows
		void shouldHaveSecondFachstelleName() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[1].name").value(fachstelle2.getName()));
		}

		@Test
		@SneakyThrows
		void shouldHaveSecondFachstelleAnschrift() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[1].anschrift").value(fachstelle2.getAnschrift()));
		}

		@Test
		@SneakyThrows
		void shouldHaveSecondFachstelleEmail() {
			var response = performRequest();
			response.andExpect(jsonPath("$.content[1].email").value(fachstelle2.getEmail()));
		}

		@SneakyThrows
		private ResultActions performRequest() {
			return mockMvc.perform(get(FachstelleController.PATH)
					.param("searchBy", searchBy)).andExpect(status().isOk());

		}
	}

}
