package de.ozgcloud.alfa.common;

import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.hateoas.Link;

import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.wiedervorlage.Wiedervorlage;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageTestFactory;

class CollectionModelBuilderTest {

	private final String HREF = "http://test";
	private final String REL = "rel";

	@Nested
	class TestBuildModel {

		@Test
		void shouldBuildModel() {
			var vorgang = VorgangHeaderTestFactory.create();

			var model = CollectionModelBuilder.fromEntities(List.of(vorgang)).buildModel();

			assertThat(model.getContent()).hasSize(1).first().usingRecursiveComparison().isEqualTo(vorgang);
		}
	}

	@Nested
	class TestAddLink {

		@Test
		void shouldAddLink() {
			var model = CollectionModelBuilder.fromEntities(List.of()).addLink(Link.of(HREF, REL)).buildModel();

			assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
		}
	}

	@Nested
	class TestIfMatch {

		@Nested
		class TestWithBooleanSupplier {

			@Test
			void shouldAddLink() {
				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> true).addLink(Link.of(HREF, REL)).buildModel();

				assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
			}

			@Test
			void shouldNotAddLink() {
				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> false).addLink(Link.of(HREF, REL)).buildModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}

		@Nested
		class TestWithPredicate {

			private final Stream<Wiedervorlage> wiedervorlageStream = Stream.of(WiedervorlageTestFactory.create());

			@Test
			void shouldAddLink() {
				var model = CollectionModelBuilder.fromEntities(wiedervorlageStream)
						.ifMatch(wiedervorlagen -> true)
						.addLink(Link.of(HREF, REL))
						.buildModel();

				assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
			}

			@Test
			void shouldNotAddLink() {
				var model = CollectionModelBuilder.fromEntities(wiedervorlageStream)
						.ifMatch(wiedervorlagen -> false)
						.addLink(Link.of(HREF, REL))
						.buildModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}
	}

}