package de.ozgcloud.alfa.historie;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserProfile;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserService;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class VorgangChangeHistoryServiceITCase {

	private static final UserId USER_1_ID = UserId.from("user-1");
	private static final String USER_1_FIRST_NAME = LoremIpsum.getInstance().getFirstName();
	private static final String USER_1_LAST_NAME = LoremIpsum.getInstance().getLastName();
	private static final String USER_1_FULL_NAME = String.format("%s %s", USER_1_FIRST_NAME, USER_1_LAST_NAME);

	private static final UserId USER_2_ID = UserId.from("user-2");
	private static final String USER_2_FIRST_NAME = LoremIpsum.getInstance().getFirstName();
	private static final String USER_2_LAST_NAME = LoremIpsum.getInstance().getLastName();
	private static final String USER_2_FULL_NAME = String.format("%s %s", USER_2_FIRST_NAME, USER_2_LAST_NAME);

	private static final String AKTENZEICHEN_1 = "ABC-1";
	private static final String AKTENZEICHEN_2 = "ABC-2";

	private final VorgangWithEingang vorgangWithEingang = VorgangWithEingangTestFactory.create();

	@Mock
	private CommandService commandService;
	@Mock
	private UserService userService;

	@InjectMocks
	private VorgangChangeHistoryService service;

	@Test
	void shouldReturnEmptyHistory() {
		givenHistorieServiceReturnsCommands(List.of());

		var history = service.createVorgangChangeHistory(vorgangWithEingang);

		assertThat(history.getStatusChangeHistory()).isEmpty();
		assertThat(history.getAktenzeichenChangeHistory()).isEmpty();
		assertThat(history.getAssignedUserChangeHistory()).isEmpty();
	}

	@Test
	void shouldReturnNonEmptyHistories() {
		givenHistorieServiceReturnsCommands(createMixedCommands());
		givenUserServiceReturnsUser1Profile();

		var history = service.createVorgangChangeHistory(vorgangWithEingang);

		assertThat(history.getStatusChangeHistory()).hasSize(1);
		assertThat(history.getAktenzeichenChangeHistory()).hasSize(1);
		assertThat(history.getAssignedUserChangeHistory()).hasSize(1);
	}

	@Test
	void shouldSetCreatedByName() {
		givenHistorieServiceReturnsCommands(List.of(new CommandFactory().statusChange(CommandOrder.VORGANG_ANNEHMEN)));

		var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

		assertThat(history.get(0).getAuthorFullName()).isEqualTo(USER_1_FULL_NAME);
	}

	@Test
	void shouldSetFinishedAt() {
		var command = new CommandFactory().statusChange(CommandOrder.VORGANG_ANNEHMEN);
		givenHistorieServiceReturnsCommands(List.of(command));

		var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

		assertThat(history.get(0).getFinishedAt()).isNotNull().isEqualTo(command.getFinishedAt());
	}

	@Nested
	class TestStatusChangeHistory {
		@Test
		void shouldSetBeforeAndAfterValues_untilAbgeschlossen() {
			givenHistorieServiceReturnsCommands(createStatusChangeCommandsUntilAbgeschlossen());

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

			assertThat(history).hasSize(6);
			checkBeforeAndAfterValues(history.get(0),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ANGENOMMEN));
			checkBeforeAndAfterValues(history.get(1),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ANGENOMMEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.IN_BEARBEITUNG));
			checkBeforeAndAfterValues(history.get(2),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.IN_BEARBEITUNG),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.BESCHIEDEN));
			checkBeforeAndAfterValues(history.get(3),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.BESCHIEDEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ABGESCHLOSSEN));
			checkBeforeAndAfterValues(history.get(4),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ABGESCHLOSSEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ZU_LOESCHEN));
			checkBeforeAndAfterValues(history.get(5),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ZU_LOESCHEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ABGESCHLOSSEN));
		}

		@Test
		void shouldSetBeforeAndAfterValues_untilVerworfen() {
			givenHistorieServiceReturnsCommands(createStatusChangeCommandsUntilVerworfen());

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

			assertThat(history).hasSize(3);
			checkBeforeAndAfterValues(history.get(0),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN));
			checkBeforeAndAfterValues(history.get(1),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ZU_LOESCHEN));
			checkBeforeAndAfterValues(history.get(2),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ZU_LOESCHEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN));
		}

		@Test
		void shouldSetBeforeAndAfterValues_untilZurueckholen() {
			givenHistorieServiceReturnsCommands(createStatusChangeCommandsUntilZurueckholen());

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

			assertThat(history).hasSize(2);
			checkBeforeAndAfterValues(history.get(0),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN));
			checkBeforeAndAfterValues(history.get(1),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN),
					StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU));
		}

		@Test
		void shouldSetOrder() {
			var command = new CommandFactory().statusChange(CommandOrder.VORGANG_ANNEHMEN);
			givenHistorieServiceReturnsCommands(List.of(command));

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getStatusChangeHistory();

			assertThat(history.get(0).getOrder()).isNotNull().isEqualTo(CommandOrder.VORGANG_ANNEHMEN.name());
		}
	}

	@Nested
	class TestAktenzeichenChangeHistory {

		@Test
		void shouldSetBeforeAndAfterValues() {
			givenHistorieServiceReturnsCommands(createAktenzeichenChangeCommands());

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getAktenzeichenChangeHistory();

			assertThat(history).hasSize(3);
			checkBeforeAndAfterValues(history.get(0), "", AKTENZEICHEN_1);
			checkBeforeAndAfterValues(history.get(1), AKTENZEICHEN_1, AKTENZEICHEN_2);
			checkBeforeAndAfterValues(history.get(2), AKTENZEICHEN_2, "");
		}

		@Test
		void shouldSetOrder() {
			var command = new CommandFactory().aktenzeichenChange(AKTENZEICHEN_1);
			givenHistorieServiceReturnsCommands(List.of(command));

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getAktenzeichenChangeHistory();

			assertThat(history.get(0).getOrder()).isNotNull().isEqualTo(CommandOrder.SET_AKTENZEICHEN.name());
		}
	}

	@Nested
	class TestAssignedUserChangeHistory {

		@Test
		void shouldSetBeforeAndAfterValues() {
			givenHistorieServiceReturnsCommands(createUserChangeCommands());
			givenUserServiceReturnsUserProfiles();

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getAssignedUserChangeHistory();

			assertThat(history).hasSize(2);
			checkBeforeAndAfterValues(history.get(0), "", USER_1_FULL_NAME);
			checkBeforeAndAfterValues(history.get(1), USER_1_FULL_NAME, USER_2_FULL_NAME);
		}

		@Test
		void shouldSetOrder() {
			var command = new CommandFactory().userChange(USER_1_ID.toString());
			givenHistorieServiceReturnsCommands(List.of(command));
			givenUserServiceReturnsUser1Profile();

			var history = service.createVorgangChangeHistory(vorgangWithEingang).getAssignedUserChangeHistory();

			assertThat(history.get(0).getOrder()).isNotNull().isEqualTo(CommandOrder.ASSIGN_USER.name());
		}
	}

	private void checkBeforeAndAfterValues(VorgangChange vorgangChange, String expectedBefore, String expectedAfter) {
		assertThat(vorgangChange.getValueBeforeChange()).isEqualTo(expectedBefore);
		assertThat(vorgangChange.getValueAfterChange()).isEqualTo(expectedAfter);
	}

	private void givenHistorieServiceReturnsCommands(List<Command> commands) {
		when(commandService.findFinishedCommands(VorgangHeaderTestFactory.ID)).thenReturn(commands.stream());
	}

	private void givenUserServiceReturnsUserProfiles() {
		givenUserServiceReturnsUser1Profile();
		givenUserServiceReturnsUser2Profile();
	}

	private void givenUserServiceReturnsUser1Profile() {
		when(userService.getById(USER_1_ID)).thenReturn(userProfile(USER_1_FIRST_NAME, USER_1_LAST_NAME));
	}

	private void givenUserServiceReturnsUser2Profile() {
		when(userService.getById(USER_2_ID)).thenReturn(userProfile(USER_2_FIRST_NAME, USER_2_LAST_NAME));
	}

	private static UserProfile userProfile(String firstName, String lastName) {
		return UserProfileTestFactory.createBuilder().firstName(firstName).lastName(lastName).build();
	}

	private List<Command> createStatusChangeCommandsUntilAbgeschlossen() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.statusChange(CommandOrder.VORGANG_ANNEHMEN),
				commandFactory.statusChange(CommandOrder.VORGANG_BEARBEITEN),
				commandFactory.statusChange(CommandOrder.VORGANG_BESCHEIDEN),
				commandFactory.statusChange(CommandOrder.VORGANG_ABSCHLIESSEN),
				commandFactory.statusChange(CommandOrder.VORGANG_ZUM_LOESCHEN_MARKIEREN),
				commandFactory.statusChange(CommandOrder.VORGANG_ABSCHLIESSEN),
				commandFactory.statusChange(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN));
	}

	private List<Command> createStatusChangeCommandsUntilVerworfen() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.statusChange(CommandOrder.VORGANG_VERWERFEN),
				commandFactory.statusChange(CommandOrder.VORGANG_ZUM_LOESCHEN_MARKIEREN),
				commandFactory.statusChange(CommandOrder.VORGANG_VERWERFEN),
				commandFactory.statusChange(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN));
	}

	private List<Command> createStatusChangeCommandsUntilZurueckholen() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.statusChange(CommandOrder.VORGANG_VERWERFEN),
				commandFactory.statusChange(CommandOrder.VORGANG_ZURUECKHOLEN));
	}

	private List<Command> createAktenzeichenChangeCommands() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.aktenzeichenChange(AKTENZEICHEN_1),
				commandFactory.aktenzeichenChange(AKTENZEICHEN_2),
				commandFactory.aktenzeichenChange(null));
	}

	private List<Command> createUserChangeCommands() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.userChange(USER_1_ID.toString()),
				commandFactory.userChange(USER_2_ID.toString()));
	}

	private List<Command> createMixedCommands() {
		var commandFactory = new CommandFactory();
		return List.of(
				commandFactory.statusChange(CommandOrder.VORGANG_ANNEHMEN),
				commandFactory.aktenzeichenChange(AKTENZEICHEN_1),
				commandFactory.userChange(USER_1_ID.toString()));
	}

	private static class CommandFactory {

		private ZonedDateTime dateTime = ZonedDateTime.now(ZoneId.of("UTC"));

		Command userChange(String assignedTo) {
			return CommandTestFactory.createBuilder()
					.order(CommandOrder.ASSIGN_USER.name())
					.body(assignedTo != null ? Map.of(AssignedUserChangeHistoryBuilder.BODY_PROPERTY_ASSIGNED_USER, assignedTo) : Map.of())
					.createdByName(USER_1_FULL_NAME)
					.finishedAt(getAndIncrementDateTime())
					.build();
		}

		Command aktenzeichenChange(String aktenzeichen) {
			return CommandTestFactory.createBuilder()
					.order(CommandOrder.SET_AKTENZEICHEN.name())
					.body(aktenzeichen != null ? Map.of(AktenzeichenChangeHistoryBuilder.BODY_PROPERTY_AKTENZEICHEN, aktenzeichen) : Map.of())
					.createdByName(USER_1_FULL_NAME)
					.finishedAt(getAndIncrementDateTime())
					.build();
		}

		Command statusChange(CommandOrder order) {
			return CommandTestFactory.createBuilder()
					.order(order.name())
					.createdByName(USER_1_FULL_NAME)
					.finishedAt(getAndIncrementDateTime())
					.build();
		}

		private ZonedDateTime getAndIncrementDateTime() {
			var current = dateTime;
			dateTime = dateTime.plusMinutes(1);
			return current;
		}
	}
}
